/* -*- Mode: C++; tab-width: 2; indent-tabs-mode: nil; c-basic-offset: 2 -*- */
/* ***** BEGIN LICENSE BLOCK *****
 * Version: NPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Netscape Public License
 * Version 1.1 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/NPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the mozCellML code.
 *
 * The Initial Developer of the Original Code is
 * The Bioengineering Institute, University of Auckland.
 * Portions created by the Initial Developer are Copyright (C) 2004
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *   Andrew Miller <ak.miller@auckland.ac.nz>
 *
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the NPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the NPL, the GPL or the LGPL.
 *
 * ***** END LICENSE BLOCK ***** */
#include "mozCellMLIntegrator.h"
#include <nsArray.h>
#include <nsIEventQueueService.h>
#include <nsIServiceManager.h>
#include "mozCellMLVariable.h"
#include "mozCellMLIPointWatcher.h"

NS_IMPL_ISUPPORTS5(mozCellMLIntegrator,
                   mozCellMLIIntegrator,
                   nsIInputStreamCallback,
                   nsIOutputStreamCallback,
                   nsIServerSocketListener,
                   nsISupportsWeakReference
                  )

mozCellMLIntegrator::mozCellMLIntegrator(
                                         nsIServerSocket* listener,
                                         nsCOMArray<mozCellMLVariable>& idifferentials,
                                         nsCOMArray<mozCellMLVariable>& isimples,
                                         nsCOMArray<mozCellMLVariable>& iconstants,
                                         nsACString& ifilename,
                                         PRUint32 ihidden_higher
                                        )
  : differentials(idifferentials),
    simples(isimples), constants(iconstants), buf_pos(0),
    hidden_higher(ihidden_higher), IntegrationRuns(NULL),
    wantcount(1), filename(ifilename)
{
  NS_ADDREF(this);
  LEAKTRACE_CREATED(mozCellMLIntegrator)
  nsCOMPtr<nsIServiceManager> servMan;
  nsCOMPtr<nsIEventQueueService> eqs = do_GetService(NS_EVENTQUEUESERVICE_CONTRACTID);
  nsCOMPtr<nsIServerSocketListener> wl = new WeakServerSocketListenerAdaptor(this);
  listener->AsyncListen(wl);
  eqs->GetSpecialEventQueue(nsIEventQueueService::
                            CURRENT_THREAD_EVENT_QUEUE,
                            getter_AddRefs(eventQueue));
  numvars = differentials.Count() + simples.Count();
  variable = new mozCellMLIVariable*[numvars];
  PRInt32 i, j;
  for (i = 0; i < simples.Count(); i++)
    variable[i] = simples[i];
  for (j = i; i < differentials.Count() + j; i++)
    variable[i] = differentials[i - j];
  max_packet = (sizeof(IntegratorReplyHeader) +
                (1 + numvars + hidden_higher) * sizeof(double));
  in_buffer = (char*)PR_Malloc(max_packet);
  transport = NULL;
}

mozCellMLIntegrator::~mozCellMLIntegrator()
{
  LEAKTRACE_DELETED(mozCellMLIntegrator)
  while (IntegrationRuns.GetSize())
    delete NS_STATIC_CAST(IntegrationRun*, IntegrationRuns.Pop());

  if (variable)
    delete [] variable;

  if (in_buffer)
    PR_Free(in_buffer);

  PR_Delete(filename.get());
}

NS_IMETHODIMP
mozCellMLIntegrator::Start(mozICellML* model,
                           double startX,
                           double stopX,
                           double resolution,
                           double abs_tolerance,
                           double rel_tolerance,
                           double max_stepsize
                          )
{
  // Compute the length of the packet...
  PRUint32 packet_len = sizeof(struct HelperCommandHeader) +
                        sizeof(struct StartIntegrationCommand) +
                        (numvars + hidden_higher + constants.Count())
                          * sizeof(double);

  if (model == nsnull)
    return NS_ERROR_FAILURE;

  // Set up the packet storage...
  struct HelperCommandHeader* hdr = (struct HelperCommandHeader*)PR_Malloc(packet_len);
  if (hdr == nsnull)
    return NS_ERROR_OUT_OF_MEMORY;
  struct StartIntegrationCommand* cmd = (struct StartIntegrationCommand*)(hdr + 1);
  double* initialValues = (double*)(cmd + 1);

  // Set up the packet and command headers...
  hdr->length = packet_len;
  hdr->cmd = START_HELPER;
  cmd->startX = startX;
  cmd->stopX = stopX;
  cmd->resolution = resolution;
  cmd->abs_tolerance = abs_tolerance;
  cmd->rel_tolerance = rel_tolerance;
  cmd->max_stepsize = max_stepsize;

  // Set up the initial values...
  PRUint32 i;
  for (i = 0; i < numvars; i++)
  {
    nsCOMPtr<mozCellMLIVariable> v;
    double val = 0;
    nsAutoString n, c;
    variable[i]->GetName(n);
    variable[i]->GetComponent(c);
    model->GetVariable(n, c, getter_AddRefs(v));
    if (v == NULL)
    {
      PR_Free(hdr);
      return NS_ERROR_FAILURE;
    }
    v->GetInitialValue(&val);
    *initialValues++ = val;
  }
  // XXX the user should be able to specify initial values for hidden higher
  //     variables.
  for (i = 0; i < hidden_higher; i++)
    *initialValues++ = 0;

  PRUint32 l = constants.Count();
  for (i = 0; i < l; i++)
  {
    nsCOMPtr<mozCellMLIVariable> v;
    double val = 0;
    nsAutoString n, c;
    constants[i]->GetName(n);
    constants[i]->GetComponent(c);
    model->GetVariable(n, c, getter_AddRefs(v));
    if (v == NULL)
    {
      PR_Free(hdr);
      return NS_ERROR_FAILURE;
    }
    v->GetInitialValue(&val);
    *initialValues++ = val;
  }

  Enqueue(hdr, packet_len);
  PR_Free(hdr);

  IntegrationRun* ir = new IntegrationRun(next_run_watchers);
  next_run_watchers.Clear();
  IntegrationRuns.PushFront(ir);

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::Suspend()
{
  struct HelperCommandHeader ich;

  ich.length = sizeof(ich);
  ich.cmd = SUSPEND_HELPER;
  Enqueue(&ich, sizeof(ich));

  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::Want()
{
  wantcount++;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::Destroy()
{
  struct HelperCommandHeader ich;

  if (--wantcount > 0)
    return NS_OK;

  ich.length = sizeof(ich);
  ich.cmd = STOP_HELPER;
  Enqueue(&ich, sizeof(ich));
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::RegisterWatcher(mozCellMLIPointWatcher* watcher)
{
  if (watcher == NULL)
    return NS_ERROR_FAILURE;
  next_run_watchers.AppendObject(NS_GetWeakReference(watcher));
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::ProcessPacket(IntegratorReply reply, void* data, PRUint32 length)
{
#ifdef DEBUG_MODE
  printf("Processing a packet %u of length %u\n", reply, length);
#endif
  if (IntegrationRuns.GetSize() == 0)
    return NS_OK;

  IntegrationRun* ir = NS_STATIC_CAST(IntegrationRun*, IntegrationRuns.Peek());

  double *ddata = (double*)data;

  for (PRUint32 i = 0, l = ir->watchers.Count();
       i < l;
       i++)
  {
    nsCOMPtr<mozCellMLIPointWatcher> watcher =
      do_QueryReferent(ir->watchers[i]);
#ifdef DEBUG_MODE
    printf("Checking watcher %u\n", i);
#endif
    if (watcher == NULL)
    {
#ifdef DEBUG_MODE
      printf("Watcher has been released.\n");
#endif
      continue;
    }

    switch (reply)
    {
    case INTEGRATION_DONE:
      watcher->Done();
      break;
    case INTEGRATION_FAILED:
    {
      nsCAutoString message;
      // XXX we shouldn't assume it is null terminated(?) although with the
      // current integrator this is guaranteed.
      if (reply == INTEGRATION_FAILED)
        message = (char*)data;
      watcher->Failed(message);
      break;
    }
    case INTEGRATION_POINT:
      watcher->PointAdded(ddata[0], numvars, variable, ddata + 1);
      break;
    case INTEGRATION_HALTED:
      watcher->Halted();
    }
  }

  if (reply == INTEGRATION_DONE ||
      reply == INTEGRATION_FAILED ||
      reply == INTEGRATION_HALTED)
  {
    IntegrationRun* ir = NS_STATIC_CAST(IntegrationRun*, IntegrationRuns.Pop());
    delete ir;
  }
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::OnSocketAccepted(
                                      nsIServerSocket* servsocket,
                                      nsISocketTransport* newsocket
                                     )
{
#ifdef DEBUG_MODE
  printf("OnSocketAccepted() called.\n");
#endif
  if (transport == NULL)
  {
    transport = newsocket;
    nsCOMPtr<nsIInputStream> input_s;
    transport->OpenInputStream(
                               nsITransport::OPEN_UNBUFFERED, 0, 0,
                               getter_AddRefs(input_s)
                              );
    input = do_QueryInterface(input_s);

    nsCOMPtr<nsIOutputStream> output_s;
    transport->OpenOutputStream(
                                nsITransport::OPEN_UNBUFFERED, 0, 0,
                                getter_AddRefs(output_s)
                               );
    output = do_QueryInterface(output_s);
    
    nsCOMPtr<nsIInputStreamCallback> wica =
      new WeakInputCallbackAdaptor(this);
    input->AsyncWait(wica, 0, 0, eventQueue);
    if (out_buffer.Length() != 0)   
    {
      awaiting_callback = PR_TRUE;
      nsCOMPtr<nsIOutputStreamCallback> woca =
        new WeakOutputCallbackAdaptor(this);
      output->AsyncWait(woca, 0, 0, eventQueue);
    }
    else
      awaiting_callback = PR_FALSE;
    OnOutputStreamReady(output);
  }
  servsocket->Close();
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::OnStopListening(nsIServerSocket* socket, PRUint32 error)
{
  return NS_OK;
}


NS_IMETHODIMP
mozCellMLIntegrator::OnInputStreamReady(nsIAsyncInputStream* ais)
{
#ifdef DEBUG_MODE
  printf("Input stream is ready.\n");
#endif
  // Read as much data as possible...
  PRUint32 actually_read;
  while (PR_TRUE)
  {
    nsresult ret = ais->Read(in_buffer + buf_pos, max_packet - buf_pos,
                             &actually_read);
    if (ret == NS_BASE_STREAM_WOULD_BLOCK)
    {
      nsCOMPtr<nsIInputStreamCallback> wica
        = new WeakInputCallbackAdaptor(this);
      ais->AsyncWait(wica, 0, 0, eventQueue);
      return NS_OK;
    }
    if (ret != NS_OK || actually_read == 0)
      return ret;
    buf_pos += actually_read;
    if (buf_pos < sizeof(IntegratorReplyHeader))
      continue;
    IntegratorReplyHeader* hdr = (IntegratorReplyHeader*)in_buffer;
    PRUint32 hl = hdr->length;
    if (hl < sizeof(IntegratorReplyHeader))
      hl = sizeof(IntegratorReplyHeader);
    if (buf_pos < hl)
      continue;
    ProcessPacket(hdr->reply, (void*)(hdr + 1),
                  hl - sizeof(IntegratorReplyHeader));
    memmove(in_buffer, in_buffer + hl, buf_pos - hdr->length);
    buf_pos -= hl;
  }
}

NS_IMETHODIMP
mozCellMLIntegrator::OnOutputStreamReady(nsIAsyncOutputStream* aos)
{
#ifdef DEBUG_MODE
  printf("Output stream is ready.\n");
#endif
  PRUint32 wrote_length;
  // XXX This is inefficient, we could use a ringbuffer etc...
  while (out_buffer.Length() != 0 &&
         aos->Write(out_buffer.get(), out_buffer.Length(), &wrote_length)
         == NS_OK)
    out_buffer = Substring(out_buffer, wrote_length);

  if (out_buffer.Length() != 0)
  {
    nsCOMPtr<nsIOutputStreamCallback> woca
      = new WeakOutputCallbackAdaptor(this);
    output->AsyncWait(woca, 0, 0, eventQueue);
  }
  else
    awaiting_callback = PR_FALSE;
  return NS_OK;
}

NS_IMETHODIMP
mozCellMLIntegrator::Enqueue(const void* ptr, PRUint32 length)
{
#ifdef DEBUG_MODE
  printf("Enqueue() called\n");
#endif
  out_buffer.Append((char*)ptr, length);
  if (transport != NULL &&
      !awaiting_callback)
  {
    awaiting_callback = PR_TRUE;
    nsCOMPtr<nsIOutputStreamCallback> woca
      = new WeakOutputCallbackAdaptor(this);
    output->AsyncWait(woca, 0, 0, eventQueue);
  }
  return NS_OK;
}

